# RC version.

#
# Standard Module Imports 
#

import time
import constants
import es

authmode = 1
updatestyle = 0
paramstyle = "?"
cache = None

# These act as modules
sql = None
db = None

# Database connection and cursor variables
conn = None
cur = None



def load():
    global cache
    
    openDatabase()
    cache = Cache()
    if not es.exists("command", "afauth"):
        es.regcmd("afauth", "afauth/handler", "Handles Stuff!")

def unload():
    closeDatabase()



def openDatabase():
    """ Open the Database and get an execution cursor. """
    global conn, cur, paramstyle, sql, db
    if constants.sqldb == "sqlite3":
        import sqlite3 as db
        import sql_sqlite3 as sql
        
        conn = db.connect(constants.sqlname)
        cur = conn.cursor()
        
    elif constants.sqldb == "mysql":
        import MySQLdb as db
        import sql_mysql as sql
        conn = db.connect(constants.sqlhost, constants.sqluser,
                          constants.sqlpass, constants.sqlname)
        cur = conn.cursor()
        
    else:
        dbg("Database not recognized: %s" % str(constants.sqldb) )
    
    
    if db.paramstyle == 'qmark':
        paramstyle = '?'
    elif db.paramstyle == 'format':
        paramstyle = '%s'
    else:
        dbg("Unknown Paramstyle")

def closeDatabase():
    """ Close the database gracefully and destroy the cursor. """
    global conn, cur
    cur = None
    if conn != None:
        try:
            conn.close()

        except:
            dbg("Connection Could Not be properly closed!")
            conn = None

def createTables():
    """ Create all tables and set some default values. """
    global cur, conn
    cur.execute(sql.CREATE_ADMINS_TABLE)
    cur.execute(sql.CREATE_GROUPS_TABLE)
    cur.execute(sql.CREATE_ADMINS_GROUP_TABLE)
    cur.execute(sql.CREATE_CONFIG_TABLE)
    cur.execute(sql.SET_TABLE_VERSION)
    cur.execute(sql.SET_TABLE_TIME )
    conn.commit()

def updateTableTime():
    """ Update when the table was last modified """
    cur.execute(sql.UPDATE_TABLE_TIME % (int(time.time() ) ) )
    conn.commit()

def getTableTime():
    """ Get when the table was last updated """
    cur.execute(sql.SELECT_TABLE_TIME)
    res = cur.fetchone()
    return int(res[1])

def dbg(msg):
    """ Placeholder for es.dbgmsg """
    pass
    #print msg


def addAdmin(name, steamid, ip, flags, immunity, password):
    """ Add an admin to the tables.  This DOES NOT update the cache. """
    cur.execute(sql.INSERT_ADMIN.replace("%s", paramstyle), (name,
                                                             steamid,
                                                             ip,
                                                             flags,
                                                             immunity,
                                                             password
                                                             )
                )
    conn.commit()
    updateTableTime()

def addGroup(name, flags):
    """ Add a group to the tables.  This DOES NOT update the cache. """
    cur.execute(sql.INSERT_GROUP.replace("%s", paramstyle), (name, flags) )
    conn.commit()
    updateTableTime()

def addAdminGroup(admin_id, group_id):
    """ Add an admin to a group.  This DOES NOT update the cache. """
    admin_id = int(admin_id); group_id = int(group_id)
    cur.execute(sql.INSERT_ADMIN_GROUP.replace("%s", paramstyle), (admin_id, group_id) )
    conn.commit()
    updateTableTime()

def delAdmin(admin_id):
    """ Delete an admin from the tables.  This DOES NOT update the cache. """
    admin_id = int(admin_id);
    cur.execute(sql.DELETE_ADMIN % admin_id)
    cur.execute(sql.DELETE_ADMIN_GROUP_BY_ADMIN % admin_id )
    conn.commit()
    
def delGroup(group_id):
    """ Delete a group from the tables.  This DOES NOT update the cache. """
    group_id = int(group_id)
    cur.execute(sql.DELETE_GROUP % group_id)
    cur.execute(sql.DELETE_ADMIN_GROUP_BY_GROUP % group_id)
    conn.commit()

def delAdminGroup(admin_id):
    """ Removes an admin from a group.  This DOES NOT update the cache. """
    admin_id = int(admin_id);
    cur.execute(sql.DELETE_ADMIN_GROUP_BY_ADMIN % admin_id )
    conn.commit()

#
# These are module functions to interact with the Cache object.
# They are all aliases of of the cache methods.
#
def update(force = 0):
    """ Update the player cache. The 'force' parameter forces an update,
    rather than checking to see if the tables have changed. """
    cache.update(force)

def getName(userid):
    """ Returns the name of the player, according to the cache, None if auth failed.  """
    return cache.getName(userid)

def getFlags(userid):
    """ Returns flags if auth is successful, and None if auth failed. """
    return cache.getFlags(userid)

def getImmunity(userid):
    """ Returns the immunity level if auth is successful, and 0 if auth failed. """
    return cache.getImmunity(userid)

def getLevel(userid):
    """ An alias of getImmunity() """
    return cache.getImmunity(userid)

def hasAdmin(userid):
    """ Checks to see if a player is in the cache. """
    return cache.hasAdmin(userid)
    
def hasFlag(userid, flag):
    """ Returns whether or not a player has a flag. """
    return cache.hasFlag(userid, flag)

def getAuthMode():
    """ Returns the current Authmode as an integer. """
    return authmode

def setAuthMode(steamid, ip, name):
    """ Sets the authmode.  Use booleans for the steamid, ip, and name parameters. """
    global authmode
    authmode = steamid | (ip<<1) | (name<<2)
    
    
class Cache(object):
    def __init__(self):
        self.update_time = 0
        self.update()
        

    def update(self, force = 0):
        """ Update the player cache. The 'force' parameter forces an update,
        rather than checking to see if the tables have changed. """

        if force == 0:
            if self.update_time == getTableTime():
                return
            else:
                dbg("Different table times, updating...")
                
        self.update_time = getTableTime()
        self.admins = {}
        self.groups = {}
        self.admins_groups = {}

        self.names = {}
        self.steamids = {}
        self.ips = {}
        self.userids = {}
        

        # Fill up the player cache
        cur.execute(sql.SELECT_ALL_PLAYERS)
        
        for each in cur.fetchall():
            self.admins[each[0]] = {}
            for i in range(len(cur.description) ):
   
                self.admins[each[0]][cur.description[i][0]] = each[i]
                #self.admins[each[0]][i] = each[i] # Numeric index

            self.admins[each[0]]['baseflags'] = self.admins[each[0]]['flags']
            self.admins[each[0]]['groupflags'] = ""
            self.names[self.admins[each[0]]['name'] ] = each[0]
            self.steamids[self.admins[each[0]]['steamid'] ] = each[0]
            self.ips[self.admins[each[0]]['ip'] ] = each[0]

        # Fill up the group cache
        cur.execute(sql.SELECT_ALL_GROUPS)

        for each in cur.fetchall():
            self.groups[each[0]] = {}

            for i in range(len(cur.description) ):
   
                self.groups[each[0]][cur.description[i][0]] = each[i]
                #self.groups[each[0]][i] = each[i] # Numeric index

        # Add the admins to the groups
        cur.execute(sql.SELECT_ADMINS_GROUPS)

        for each in cur.fetchall():
            self.admins[each[0]]['groupflags'] += self.groups[each[1]]['flags']

        # Remove duplicate flags
        for each in self.admins:
            
            self.admins[each]['flags'] = "".join(sorted(set(self.admins[each]['baseflags'] +
                                                            self.admins[each]['groupflags']) ) )
                                                            
          
            self.admins[each]['baseflags'] = "".join(sorted(self.admins[each]['baseflags']))
            self.admins[each]['groupflags'] = "".join(sorted(self.admins[each]['groupflags']))

    def getName(self, userid):
        """ Returns the name of the player, according to the cache, None if auth failed.  """
        adminid = self.getAdmin(userid)
        if adminid != 0:
            return self.admins[adminid]['name']
        else:
            return None
            
    def getFlags(self, userid):
        """ Returns flags if auth is successful, and None if auth failed. """
        adminid = self.getAdmin(userid)
        if adminid != 0:
            return self.admins[adminid]['flags']
        else:
            return None

    def getImmunity(self, userid):
        """ Returns the immunity level if auth is successful, and 0 if auth failed. """
        adminid = self.getAdmin(userid)
        if adminid != 0:
            return self.admins[adminid]['immunity']
        else:
            return 0

    def getLevel(self, userid):
        """ An alias of getImmunity() """
        return self.getImmunity(userid)

    def hasAdmin(self, userid):
        """ Public version of getAdmin().  Use this to see if a user is in the cache. """
        adminid = self.getAdmin(userid)
        if adminid != 0:
            return True
        else:
            return False

    def hasFlag(self, userid, flag):
        """ Returns whether or not a player has a flag. """
        adminid = self.getAdmin(userid)
        if adminid != 0:
            if flag in self.admins[adminid]['flags']:             
                return True
        return False
        
    def getAdmin(self, userid):
        """ This is a private function to get the adminid. """
        global authmode
        adminset = set()
        adminid = 0
        
        userid = str(userid)
        
        if userid in self.userids:
            return self.userids[userid]
        
        if authmode & 1:
            steamid = es.createplayerlist(userid)[int(userid)]["steamid"]
            if steamid in self.steamids:
                adminset.add(self.steamids[steamid])
            else:
                dbg("Failed steamid check")
                return 0
        if authmode & 2:
            ip = es.createplayerlist(userid)[int(userid)]["address"]
            ip = ip.split(":")[0]
            if ip in self.ips:
                adminset.add(self.ips[ip])
            else:
                dbg("Failed ip check")
                return 0
        if authmode & 4:
            name = es.createplayerlist(userid)[int(userid)]["name"]
            if name in self.names:
                adminset.add(self.names[name])
            else:
                dbg("Failed name check")
                return 0
            
        if len(adminset) == 1:
            adminid = adminset.pop()
        else:
            dbg("Multiple Choices")
            return 0

        if len(self.admins[adminid]['password']):
            password = es.queryclientvar(userid, constants.passwordvar)
            if password != self.admins[adminid]['password']:
                dbg("Bad Password")
                return 0
                
        self.userids[userid] = adminid
        return adminid

        
if __name__ == "__main__":
    load()

#
# Events
#

# Mode 0 means that the cache is refreshed on player join.
def player_connect(ev):
    global updatestyle, cache, cur
    if updatestyle == 0 and cur != None:
        cache.update()

# Mode 1 means that the cache is refreshed on map change.
def es_map_start(ev):
    global updatestyle, cache, cur
    if updatestyle == 1 and cur != None:
        cache.update()

def handler():
    arg = []
    args = es.getargs()
    argc = es.getargc()
    tell = es.dbgmsg
    
    for x in range(argc ):
        arg.append(es.getargv(x) )
    
    if argc == 1: # No command specified
        tell(0, "Available commands:")
        tell(0, "    list   - Lists all admins.")
        tell(0, "    add    - Add admins and groups.")
        tell(0, "    join   - Adds an admin to a group.")
        tell(0, "    del    - Deletes an admin or group. ")
        tell(0, "    create - Initializes the database.")
        tell(0, "    update - Forces the cache to update.")
        
        return
    else:
        if arg[1].lower() == "list":
            tell(0, "Listing %s admin(s)" % len(cache.admins) )
            tell(0, "<admin_id: name, steamid, ip, baseflags, groupflags, immunity, password>")
            for each in cache.admins:
                tell(0, "%3i: '%s', '%s', '%s', '%s', '%s', '%s', '%s'" %
                     (cache.admins[each]['id'],  cache.admins[each]['name'],
                      cache.admins[each]['steamid'], cache.admins[each]['ip'],
                      cache.admins[each]['baseflags'], cache.admins[each]['groupflags'],
                      cache.admins[each]['immunity'], cache.admins[each]['password']
                      )
                     )
            tell(0, "")
            tell(0, "Listing %s group(s)" % len(cache.groups) )
            tell(0, "<group_id: name, flags>")
            for each in cache.groups:
                tell(0, "%3i: '%s', '%s'" %
                     (cache.groups[each]['id'],  cache.groups[each]['name'],
                      cache.groups[each]['flags']
                      )
                     )
        elif arg[1].lower() == "update":
            update(1)
            tell(0, "Updated cache with %s player(s) and %s group(s)." %
                 (len(cache.admins), len(cache.groups)) )

        elif arg[1].lower() == "create":
            createTables()
            tell(0, "Tables have been created.")
        elif arg[1].lower() == "add":
            if argc > 2 and arg[2].lower() == "admin":
                if argc == 9:
                    addAdmin(arg[3], arg[4], arg[5], arg[6], arg[7], arg[8])
                    tell(0, "Added Admin")
                else:
                    tell(0, "Add Admin sytnax: ")
                    tell(0, "    afauth add admin \"name\" \"steamid\" \"ipaddress\" \"flags\" \"immunity\" \"password\"")
                    tell(0, "")
                    tell(0, "Example: afauth add admin \"Billy\" \"STEAM_0:0:1234\" \"\" \"abc\" \"99\" \"\" ")
                    tell(0, "")
                    tell(0, "Note: You must include every parameter, even if you do not plan to use it.")
                    tell(0, "Note: You can use blank strings (\"\") for fields that will not be used.")
            elif argc > 2 and arg[2].lower() == "group":
                if argc == 5:
                    addGroup(arg[3], arg[4])
                    tell(0, "Added Group")
                else:
                    tell(0, "Add Group sytnax: ")
                    tell(0, "    afauth add group \"name\" \"flags\"")
                    tell(0, "")
                    tell(0, "Example: afauth add group \"Members\" \"abc\" ")
            else:
                tell(0, "Adding admins and groups: ")
                tell(0, "    admin - Adds an admin to the database.")
                tell(0, "    group - Adds a group to the database.")
        elif arg[1].lower() == "join":
            if argc == 4:
                addAdminGroup(arg[2], arg[3])
                tell(0, "Admin joined Group"
            else:
                tell(0, "Join syntax: ")
                tell(0, "    afauth join \"admin_id\" \"group_id\" ")
                tell(0, "")
                tell(0, "Note: you can get the admin and group id's from the list command.")
        elif arg[1].lower() == "del":
            if argc > 2 and arg[2].lower() == "admin":
                if argc == 4:
                    delAdmin(arg[3])
                    tell(0, "Deleted Admin")
                else:
                    tell(0, "Delete Admin sytnax: ")
                    tell(0, "    afauth del admin \"admin_id\"")
                    tell(0, "")
                    tell(0, "Note: you can get the admin and group id's from the list command.")
            elif argc > 2 and arg[2].lower() == "group":
                if argc == 4:
                    delGroup(arg[3])
                    tell(0, "Deleted Group")
                else:
                    tell(0, "Delete Admin sytnax: ")
                    tell(0, "    afauth del group \"group_id\"")
                    tell(0, "")
                    tell(0, "Note: you can get the admin and group id's from the list command.")
            else:
                tell(0, "Deleting admins and groups: ")
                tell(0, "    admin - Deletes an admin from the database.")
                tell(0, "    group - Deletes a group from the database.")

    
